/* asm_compress_block.S -- compress_block(), for speed on i386/i586/i686.
 * Copyright 2002 BitWagon Software LLC.  All rights reserved.
 * This is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 2, see the file COPYING.
 */

#define t1   %edx  /* overlaps  len  */
#define len  %edx  /* overlaps  t1  */
#define bbuf %ebx
#define flg  %ebp
#define lxp  %esi
#define code %edi

#define l_buf inbuf

/* Inline send_bits().  Register bbuf is the bit accumulator.  Register
   %ecx= negative number of bits left in outbuf.  Thus the low 5 bits of %ecx
   are the shift count for the next deposit, a change in bit 5 signals
   that a full word is ready to be stored into outbuf, %ecx>>5 is the
   word number (backward from the hi address end) to store, and the sign
   bit changes when the outbuf is full.  [4 uses for the same register.]
   The fastest way to detect a change in bit 5 is to replicate the code,
   know which phase is current, and look for the next phase.
*/
#define send_bits(acc,len,jxx,this,that) \
	movl acc,(%esp); /* in case straddle */\
	shll %cl,acc; addl len,%ecx; \
	orl  acc,bbuf; testb $1<<5,%cl; jxx this; /* word not full */\
	movl %ecx,%eax; jmp S##that; /* change phase */\
	.align 1<<4; \
S##this: call straddle /* start next word by shifting saved acc */\
 /*this:*/

/* offsets in ct_data */
Code = 0
Len  = 2

OUTBUFSIZ = 16384  # gzip.h
END_BLOCK = 256  # trees.c
LITERALS = 256  # trees.c

	.extern base_dist    # int[]
	.extern base_length  # int[]
	.extern bi_align     # subroutine to 32bit align bi_buf, bi_valid
	.extern bi_buf       # unsigned bit buffer
	.extern bi_valid     # int number of bits in bi_buf
	.extern d_buf        # unsigned short[]
	.extern dist_code    # unsigned char[]
	.extern extra_dbits  # int[]
	.extern extra_lbits  # int[]
	.extern flag_buf     # unsigned[]
	.extern flush_outbuf # subroutine
	.extern inbuf        # unsigned char[]
	.extern last_lit     # unsigned
	.extern length_code  # unsigned char[]
	.extern outbuf       # unsigned[]
	.extern outcnt       # int

/* void compress_block(ct_data const *const ltree, ct_data const *const dtree) */
compress_block: .globl compress_block; .type compress_block,@function
	pushl %ebp; pushl %ebx; pushl %esi; pushl %edi

	movl last_lit,%eax
	movl $ l_buf,lxp
	addl lxp,%eax
	pushl $ flag_buf  # fxp_next
	pushl %eax  # lxp_last
	pushl $ d_buf  # dxp
	subl $2*4,%esp  # temp for straddle, xbits

/* %esp displacement of arguments */
ltree = 5*4 + 4*4 + 1*4
dtree = 5*4 + 4*4 + 2*4

/* %esp displacement of locals */
fxp_next = 4*4
lxp_last = 3*4
dxp      = 2*4
xbits    = 1*4

	call bi_align
	movl outcnt,%ecx; subl flg,flg  # cheat (use 1<<31 if using 'adcl')
	movl bi_buf,bbuf
	leal -OUTBUFSIZ<<3(,%ecx,8),%ecx
	addl bi_valid,%ecx

	testb $1<<5,%cl; jz loop0
	jmp loop1

	.align 1<<3
get_flag0:
	movl fxp_next(%esp),t1
	movl (t1),flg; leal 4(t1),t1
	addl flg,flg; leal 1(flg),flg  # adcl flg,flg  is 8 cycles on Pentium4
	movl t1,fxp_next(%esp); jmp got_flag0

	.align 1<<4
loop0:
	movzbl (lxp),code; leal 1(lxp),lxp  # lc
	addl flg,flg; jz get_flag0  # Carry also set if Zero
got_flag0:
	movl ltree(%esp),t1; jc no_lit0
	movzwl    (t1,code,4),%eax  # Code
	movzwl Len(t1,code,4),len
	send_bits(%eax,len, jz,bot0,bot1)
bot0:
	cmpl lxp,lxp_last(%esp); ja loop0
	jmp end0

no_lit0:
	movl code,xbits(%esp)  # save lc
	movzbl length_code(code),code
	movzwl Code+4*(1+LITERALS)(t1,code,4),%eax
	movzwl Len +4*(1+LITERALS)(t1,code,4),len
	send_bits(%eax,len, jz,lenx0,lenx1)
lenx0:
	movl extra_lbits(,code,4),len  # extra
	testl len,len; je no_xl0
	movl xbits(%esp),%eax  # lc
	subl base_length(,code,4),%eax  # lc - base_length[code]
	send_bits(%eax,len, jz,no_xl0,no_xl1)
no_xl0:
	movl dxp(%esp),t1
	movzwl (t1),%eax; addl $2,t1  # %eax= dist
	movl t1,dxp(%esp)
	movl $256,len; movl %eax,xbits(%esp)  # save dist
	cmpl len,%eax; jb dshort0
	shrl $7,%eax; addl len,%eax
dshort0:
	movl dtree(%esp),t1
	movzbl dist_code(%eax),code
	movzwl    (t1,code,4),%eax  # Code
	movzwl Len(t1,code,4),len
	send_bits(%eax,len, jz,distx0,distx1)
distx0:
	movl extra_dbits(,code,4),len  # extra
	testl len,len; je no_xd0
	movl xbits(%esp),%eax  # dist
	subl base_dist(,code,4),%eax  # dist - base_dist[code]
	send_bits(%eax,len, jz,no_xd0,no_xd1)
no_xd0:
	cmpl lxp,lxp_last(%esp); ja loop0
end0:
	movl ltree(%esp),t1
	movzwl Code+4*END_BLOCK(t1),%eax
	movzwl Len +4*END_BLOCK(t1),len
	send_bits(%eax,len, jz,done0,done1)
done0:
	jmp cb_done

get_flag1:
	movl fxp_next(%esp),t1
	movl (t1),flg; leal 4(t1),t1
	addl flg,flg; leal 1(flg),flg  # adcl flg,flg  is 8 cycles on Pentium4
	movl t1,fxp_next(%esp); jmp got_flag1

	.align 1<<4
loop1:
	movzbl (lxp),code; leal 1(lxp),lxp  # lc
	addl flg,flg; jz get_flag1  # Carry also set if Zero
got_flag1:
	movl ltree(%esp),t1; jc no_lit1
	movzwl    (t1,code,4),%eax  # Code
	movzwl Len(t1,code,4),len
	send_bits(%eax,len,jnz,bot1,bot0)
bot1:
	cmpl lxp,lxp_last(%esp); ja loop1
	jmp end1

no_lit1:
	movl code,xbits(%esp)  # save lc
	movzbl length_code(code),code
	movzwl Code+4*(1+LITERALS)(t1,code,4),%eax
	movzwl Len +4*(1+LITERALS)(t1,code,4),len
	send_bits(%eax,len,jnz,lenx1,lenx0)
lenx1:
	movl extra_lbits(,code,4),len  # extra
	testl len,len; je no_xl1
	movl xbits(%esp),%eax  # lc
	subl base_length(,code,4),%eax  # lc - base_length[code]
	send_bits(%eax,len,jnz,no_xl1,no_xl0)
no_xl1:
	movl dxp(%esp),t1
	movzwl (t1),%eax; addl $2,t1  # %eax= dist
	movl t1,dxp(%esp)
	movl $256,len; movl %eax,xbits(%esp)  # save dist
	cmpl len,%eax; jb dshort1
	shrl $7,%eax; addl len,%eax
dshort1:
	movl dtree(%esp),t1
	movzbl dist_code(%eax),code
	movzwl    (t1,code,4),%eax  # Code
	movzwl Len(t1,code,4),len
	send_bits(%eax,len,jnz,distx1,distx0)
distx1:
	movl extra_dbits(,code,4),len  # extra
	testl len,len; je no_xd1
	movl xbits(%esp),%eax  # dist
	subl base_dist(,code,4),%eax  # dist - base_dist[code]
	send_bits(%eax,len,jnz,no_xd1,no_xd0)
no_xd1:
	cmpl lxp,lxp_last(%esp); ja loop1
end1:
	movl ltree(%esp),t1
	movzwl Code+4*END_BLOCK(t1),%eax
	movzwl Len +4*END_BLOCK(t1),len
	send_bits(%eax,len,jnz,done1,done0)
done1:

cb_done:
	call xupdate
	addl $5*4,%esp
	popl %edi; popl %esi; popl %ebx; popl %ebp
	ret

	.align 1<<5
straddle: .type straddle,@function  # could be inlined; In: %eax= %ecx; len < 32
	sarl $5,                      %eax;    subl %ecx,len
	movl bbuf,outbuf+OUTBUFSIZ-4(,%eax,4); andl $~(~0<<5),len
/*	cmpl bbuf,outbuf+OUTBUFSIZ  (,%eax,4)  # tend PentiumPlain cache*/
/* In this case, 'tending' becomes 'sweeping', and we don't want that. */
	movl 4(%esp),bbuf  # unstraddled bits
	movl %ecx,%eax; movl len,%ecx
	shrl %cl,bbuf; testl %eax,%eax  # shift relies on input len < 32
	movl %eax,%ecx; jns obuf_edge
	ret
obuf_edge:
	call xupdate
	pushl %ecx; call flush_outbuf  # %eax= - number_of_bits_flushed
	popl %ecx; addl %eax,%ecx  # %ecx= new shift count
	ret

xupdate: .type xupdate,@function  # re-establish bi_buf, bi_valid, outcnt
	movl %ecx,%eax; andl $~(~0<<5),%ecx
	sarl $5,%eax;        movl %ecx,bi_valid
	leal OUTBUFSIZ(,%eax,4),%eax; movl bbuf,bi_buf
	movl %eax,outcnt
	ret

