/* asm_inflate.S -- inflate_codes(), for speed on i386/i586/i686.
 * Copyright 2002 BitWagon Software LLC.  All rights reserved.
 * This is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 2, see the file COPYING.
 */

/* k = Negative of number of bits remaining in input buffer */
#define k %ebx

/* Use unaligned fetch to avoid dedicating a register for the bit buffer 'b'.
   This incurs a 2-cycle penalty on PentiumPlain/MMX, and no penalty (!)
   on Intel "dynamic execution" processors ("P6" family: PentiumPro, II,
   III, Celeron) as long as the access does not straddle a cache line
   boundary (32 bytes).  The penalty remains 2 cycles for straddling
   a cache line boundary.

   PentiumPlain/MMX ("P5" family and relatives) does not allocate a cache
   line on a data write miss.  To prevent 32 misses in a row when writing
   consecutive bytes: read any byte in the line before writing.  A read
   miss does allocate a new cache line.  It is faster to read every byte
   in turn, rather than trying to keep track to read just one.  
*/
	.extern window     # output buffer, and look-back array
	.extern outcnt     # actual wp
	.extern mask_bits  # array of 32-bit masks ~(~0<<j), 0<=j<=16
	.extern in_end     # pointer to one beyond end of input buffer
	.extern ha         # pointer to array of struct huft
	.extern getwrap    # fn to refill the input buffer, and adjust k
	.extern flush_window  # fn to write the output buffer
	.extern crc_32_tab # array of 32-bit partial products
	.extern crc        # 32-bit running cyclic redundancy checksum

/* In: %edx= n.  Out: %eax= bits; trashes %ecx and %edx */
#define M_getbits \
1:\
	              movl k,%eax;\
	sarl $3,%eax; movl k,%ecx;\
	addl in_end,%eax; andl $7,%ecx;\
	addl %edx,k; jle 0f;\
	subl %edx,k; pushl %edx; call getwrap; popl %edx; jmp 1b;\
0:\
	movl (%eax),%eax;\
	shrl %cl,%eax; movl mask_bits(,%edx,4),%edx;\
	andl %edx,%eax

/* Same as M_getbits except k not changed [unless input wraps around] */
#define M_peekbits \
1:\
	              movl k,%eax;\
	sarl $3,%eax; movl k,%ecx;\
	addl in_end,%eax; andl $7,%ecx;\
	addl %edx,k; jle 0f;\
	subl %edx,k; pushl %edx; call getwrap; popl %edx; jmp 1b;\
0:\
	movl (%eax),%eax; subl %edx,k;\
	shrl %cl,%eax; movl mask_bits(,%edx,4),%edx;\
	andl %edx,%eax

/* Offsets within struct huft; shifting depends on little endian */
v_e = 0
v_b = 1
v_n = 2
v_t = 2

WSIZE = 1<<15
#define slide window
#define wp outcnt

#define bt %ecx  /* huft.b */
#define d %esi  /* overlaps  t */
#define e %edx  /* overlaps  n */
#define t %esi  /* overlaps  d */
#define w %edi
#define n %edx  /* overlaps  e ; also on stack while computing  d */
#define nt %esi  /* overlaps  t , d ;  another place for  n */
#define r_crc %ebp  /* also uses %ecx and %cl when non-literal */

tl = (3+1+0)*4
td = (3+1+1)*4
bl = (3+1+2)*4
bd = (3+1+3)*4


MOVZBL = (0xB6<<8) | 0x0F

inflate_codes: .globl inflate_codes  # status = inflate_codes(tl, td, bl, bd)
	pushl %ebp; pushl %edi; pushl %esi

/* This could be done just once by an init function, but forgetting to call
   that function doubles the cost of infl_copy on i686.  So spend 100 cycles.
*/
	pushfl; popl %eax; movl %eax,%ecx  # two copies of flags
	xorl $1<<21,%eax; pushl %eax; popfl  # try to change ID bit
	pushfl; popl %eax; xorl %eax,%ecx; je noCPUID  # cannot be changed
	pushl %ebx
	xorl %eax,%eax; cpuid; cmpl $1,%eax; jc noFEAT
	pushl $1; popl %eax; cpuid; testw %dx,%dx; jns noCMOV
/* is P6 or above; use 'movzbl' to prevent partial register stall */
	movl $(0x8B<<24)|(0300<<16)|MOVZBL,%eax; movl %eax,  modcopy
	movl $0344,%eax;                         movb  %al,4+modcopy
	xorl %eax,%eax; cpuid  # serialize changes to future instructions
noCMOV:
noFEAT:
	popl %ebx
noCPUID:

	movl crc,r_crc
	movl wp,w; jmp infl_main

	.align 1<<4
literal:  # In: %eax= .v_n
	movb %al,slide(w); xorl r_crc,%eax  # P6 no stall: %eax already 32 bits
	shrl $8,r_crc;     andl $0xff,%eax
	addl $1,w;  xorl crc_32_tab(,%eax,4),r_crc
	cmpl $ WSIZE,w; jl infl_main; call full
infl_main:
	movb slide(w),%al  # tend PentiumPlain cache
	movl tl(%esp),t; movl bl(%esp),%edx; jmp enter1

	.align 1<<3
loop1:  # In: %eax = .v_n === .v_t
	cmpl $99 - 16,e; mov ha,t; je infl_err
	leal (t,%eax,8),t
enter1:
	M_peekbits
	movl $0xff,e; movl (t,%eax,4),%eax
	movzbl %ah,bt  # v_b
	andl  %eax,e   # v_e
	addl bt,k
	shrl $16,%eax  # v_t === v_n
	subl $16,e; jg loop1

	addl $16,e; jnc literal
	cmpl $15,e; je done

	movl %eax,nt; M_getbits

	addl nt,%eax; movl td(%esp),t  # beware nt === t
	              movl bd(%esp),%edx
	pushl %eax; jmp enter2  # save n  [%edx is scratch]

infl_err:
	pushl $1; jmp bye
done:
	pushl $0  # success
bye:
	movl w,wp
	movl r_crc,crc
	popl %eax  # status code
	popl %esi; popl %edi; popl %ebp
	ret

	.align 1<<3
loop2:
	cmpl $99 - 16,e; movl ha,t; je infl_err
	leal (t,%eax,8),t
enter2:
	M_peekbits
	movl $0xff,e; movl (t,%eax,4),%eax
	movzbl %ah,bt  # v_b
	andl  %eax,e   # v_e
	addl bt,k
	shrl $16,%eax  # v_t === v_n
	subl $16,e; jg loop2

	addl $16,e; movl w,d
	subl %eax,d; M_getbits
	popl n  # recover n
	subl %eax,d

/* Use crc in %ecx and %cl, to avoid P6 partial register stall.
   Use WSIZE -1 in %ebp to save space in infl_copy in .data.
*/
#define wmsk %ebp
	movl r_crc,%ecx; movl $ WSIZE -1,wmsk
	andl wmsk,d; jmp infl_copy

wrap:
	pushl n; movl %ecx,r_crc; call full
	popl n; movl r_crc,%ecx
	movb slide,%al; movl $ WSIZE -1,wmsk  # tend PentiumPlain cache
	jmp more

full:
	movl w,wp; subl w,w
	pushl $1; call flush_window; addl $4,%esp
	ret

	.data; .align 1<<6
infl_copy:
	movb slide(d),%al; addl $1,d
	movb %al,slide(w); xorb %cl,%al
	shrl $8,%ecx
modcopy:              andl $0xff,%eax  # P6: "movzbl %al,%eax; movl %esp,%esp"
	addl $1,w; andl wmsk,d
	xorl crc_32_tab(,%eax,4),%ecx
	cmpl wmsk,w; movb slide(w),%al  # tend PentiumPlain cache
	ja wrap
more:
	subl $1,n; jg infl_copy
	movl %ecx,r_crc; jmp infl_main

	.align 1<<6
