/* inflate.c -- Not copyrighted 1992 by Mark Adler
   version c10p1, 10 January 1993 */

/* You can do whatever you like with this source file, though I would
   prefer that if you modify it and redistribute it that you include
   comments to that effect with your name and the date.  Thank you.
   [The history has been moved to the file ChangeLog.]
 */

/*
   2002-05-09  John Reiser  <jreiser@BitWagon.com>
   gcc global register variable k for 32-bit i386;
   remove bitbuffer b, replace with unaligned fetch from inbuf[];
   getbits(), peekbits(), getwrap() replace NEEDBITS and DUMPBITS;
   inline updcrc(); revise inflate_codes() for dataflow.
   12% speedup overall on PentiumPlain/MMX  [elapsed time "gzip -d -t"].
   31% speedup overall on PentiumPlain/MMX using asm_inflate.S;
   25% speedup overall on i686 using asm_inflate.S.

   REMEMBER TO USE  -fomit-frame-pointer  WHEN COMPILING THIS FILE.
   The C language version of inflate_codes() benefits from the extra register.
   With asm_inflate.S, the benefit might not be worth the build complexity.

   2002-05-05  John Reiser  <jreiser@BitWagon.com>
   Stack of struct huft via realloc(), saving space [cache] by using indices
   instead of pointers.  ha could be a static array, but the bounds
   are problematic; maximal observed bound was 1014.  'const' more.
   Correct comment about argument to NEEDBITS vs. DUMPBITS.
*/

unsigned const mask_bits[] = { /* 32 bits each */
    0x0000,
    0x0001, 0x0003, 0x0007, 0x000f, 0x001f, 0x003f, 0x007f, 0x00ff,
    0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff, 0xffff
};

register int k asm("%ebx");  /* 0 - number_of_bits_remaining_before(in_end) */

extern unsigned char inbuf[];
extern unsigned insize; /* valid bytes in inbuf */
extern unsigned inptr;  /* index of next byte to be processed in inbuf */
/*static*/ unsigned char *in_end;  /* asm_inflate.S */

#include <stdio.h>

/*static*/  /* asm_inflate.S */
void
getwrap()
{
	((unsigned *)inbuf)[0] = ((unsigned *)in_end)[-1];  /* WRAP_IN */
	fill_inbuf(0);  /* NOTE: returns (and removes) 1st byte */
	in_end = &inbuf[insize];
	k -= (1+ insize - inptr)<<3;  /* 1+: put back 1st byte, too */
}

static /* not inline */
unsigned
getbits(int const n)  /* remove and return next  n  bits of input */
{
	for (;;) {
		unsigned const v = mask_bits[n] &
			((*(unsigned *)&in_end[k>>3]) >> (7&k));
		if (0>=(k += n)) {
			return v;
		}
		k -= n;
		getwrap();
	}
}

/* gcc-3.1 -O2 generates wrong code for peekbits() by hoisting "(k + n)"
   out of the loop.  http://gcc.gnu.org/gnats.html bug optimization/6722
   Use gcc-2.95.* or gcc-2.96.* instead.
*/
static inline
unsigned
peekbits(int const n)  /* return next  n  bits of input, but do not remove */
{
	for (;;) {
		unsigned const v = mask_bits[n] &
			((*(unsigned *)&in_end[k>>3]) >> (7&k));
		if (0>=(k + n)) {
			return v;
		}
		getwrap();
	}
}

/*
   Inflate deflated (PKZIP's method 8 compressed) data.  The compression
   method searches for as much of the current string of bytes (up to a
   length of 258) in the previous 32K bytes.  If it doesn't find any
   matches (of at least length 3), it codes the next byte.  Otherwise, it
   codes the length of the matched string and its distance backwards from
   the current position.  There is a single Huffman code that codes both
   single bytes (called "literals") and match lengths.  A second Huffman
   code codes the distance information, which follows a length code.  Each
   length or distance code actually represents a base value and a number
   of "extra" (sometimes zero) bits to get to add to the base value.  At
   the end of each deflated block is a special end-of-block (EOB) literal/
   length code.  The decoding process is basically: get a literal/length
   code; if EOB then done; if a literal, emit the decoded byte; if a
   length then get the distance and emit the referred-to bytes from the
   sliding window of previously emitted data.

   There are (currently) three kinds of inflate blocks: stored, fixed, and
   dynamic.  The compressor deals with some chunk of data at a time, and
   decides which method to use on a chunk-by-chunk basis.  A chunk might
   typically be 32K or 64K.  If the chunk is uncompressible, then the
   "stored" method is used.  In this case, the bytes are simply stored as
   is, eight bits per byte, with none of the above coding.  The bytes are
   preceded by a count, since there is no longer an EOB code.

   If the data is compressible, then either the fixed or dynamic methods
   are used.  In the dynamic method, the compressed data is preceded by
   an encoding of the literal/length and distance Huffman codes that are
   to be used to decode this block.  The representation is itself Huffman
   coded, and so is preceded by a description of that code.  These code
   descriptions take up a little space, and so for small blocks, there is
   a predefined set of codes, called the fixed codes.  The fixed method is
   used if the block codes up smaller that way (usually for quite small
   chunks), otherwise the dynamic method is used.  In the latter case, the
   codes are customized to the probabilities in the current block, and so
   can code it much better than the pre-determined fixed codes.
 
   The Huffman codes themselves are decoded using a mutli-level table
   lookup, in order to maximize the speed of decoding plus the speed of
   building the decoding tables.  See the comments below that precede the
   lbits and dbits tuning parameters.
 */


/*
   Notes beyond the 1.93a appnote.txt:

   1. Distance pointers never point before the beginning of the output
      stream.
   2. Distance pointers can point back across blocks, up to 32k away.
   3. There is an implied maximum of 7 bits for the bit length table and
      15 bits for the actual data.
   4. If only one code exists, then it is encoded using one bit.  (Zero
      would be more efficient, but perhaps a little confusing.)  If two
      codes exist, they are coded using one bit each (0 and 1).
   5. There is no way of sending zero distance codes--a dummy must be
      sent if there are none.  (History: a pre 2.0 version of PKZIP would
      store blocks with no distance codes, but this was discovered to be
      too harsh a criterion.)  Valid only for 1.93a.  2.04c does allow
      zero distance codes, which is sent as one code of zero bits in
      length.
   6. There are up to 286 literal/length codes.  Code 256 represents the
      end-of-block.  Note however that the static length tree defines
      288 codes just to fill out the Huffman codes.  Codes 286 and 287
      cannot be used though, since there is no length base or extra bits
      defined for them.  Similarly, there are up to 30 distance codes.
      However, static trees define 32 codes (all 5 bits) to fill out the
      Huffman codes, but the last two had better not show up in the data.
   7. Unzip can check dynamic Huffman blocks for complete code sets.
      The exception is that a single code would not be complete (see #4).
   8. The five bits following the block type is really the number of
      literal codes sent minus 257.
   9. Length codes 8,16,16 are interpreted as 13 length codes of 8 bits
      (1+6+6).  Therefore, to output three times the length, you output
      three codes (1+1+1), whereas to output four times the same length,
      you only need two codes (1+3).  Hmm.
  10. In the tree reconstruction algorithm, Code = Code + Increment
      only if BitLength(i) is not zero.  (Pretty obvious.)
  11. Correction: 4 Bits: # of Bit Length codes - 4     (4 - 19)
  12. Note: length code 284 can represent 227-258, but length code 285
      really is 258.  The last length deserves its own, short code
      since it gets used a lot in very redundant files.  The length
      258 is special since 258 - 3 (the min match length) is 255.
  13. The literal/length and distance code bit lengths are read as a
      single stream of lengths.  It is possible (and advantageous) for
      a repeat code (16, 17, or 18) to go across the boundary between
      the two sets of lengths.
 */

#ifdef RCSID
static char rcsid[] = "$Id: inflate.c,v 1.1 2002/05/30 02:56:24 jreiser Exp jreiser $";
#endif

#include <config.h>
#include "tailor.h"

#if defined STDC_HEADERS || defined HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#include "gzip.h"
#define slide window
extern ulg crc;
extern ulg const crc_32_tab[];
extern ulg updcrc(uch *s, unsigned n);

/* Huffman code lookup table entry--this entry is four bytes for machines
   that have 16-bit pointers (e.g. PC's in the small or medium model).
   Valid extra bits are 0..13.  e == 15 is EOB (end of block), e == 16
   means that v is a literal, 16 < e < 32 means that v is a pointer to
   the next table, which codes e - 16 bits, and lastly e == 99 indicates
   an unused code.  If a code with e == 99 is looked up, this implies an
   error in the data. */
typedef unsigned short huft_ndx;  /* index>>1 */
struct huft {
  uch e;                /* number of extra bits or operation */
  uch b;                /* number of bits in this code or subcode */
  union {
    ush n;              /* literal, length base, or distance base */
    huft_ndx t;         /* index>>1 of next level of table */
  } v;
};


/* Function prototypes */
int huft_build OF((unsigned const *, unsigned, unsigned, ush const *, ush const *,
                   huft_ndx *, int *));
int inflate_codes OF((struct huft const *, struct huft const *, int, int));
int inflate_stored OF((void));
int inflate_fixed OF((void));
int inflate_dynamic OF((void));
int inflate_block OF((int *));
int inflate OF((void));


/* The inflate algorithm uses a sliding 32K byte window on the uncompressed
   stream to find repeated byte strings.  This is implemented here as a
   circular buffer.  The index is updated simply by incrementing and then
   and'ing with 0x7fff (32K-1). */
/* It is left to other modules to supply the 32K area.  It is assumed
   to be usable as if it were declared "uch slide[32768];" or as just
   "uch *slide;" and then malloc'ed in the latter case.  The definition
   must be in unzip.h, included above. */
/* unsigned wp;             current position in slide */
#define wp outcnt
#define flush_output(w) (wp=(w),flush_window(1))

/* Tables for deflate from PKZIP's appnote.txt. */
static unsigned const border[] = { /* Order of the bit length code lengths */
        16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};
static ush const cplens[] = {   /* Copy lengths for literal codes 257..285 */
        3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
        35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0};
        /* note: see note #13 above about the 258 in this list. */
static ush const cplext[] = {   /* Extra bits for literal codes 257..285 */
        0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2,
        3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0, 99, 99}; /* 99==invalid */
static ush const cpdist[] = {   /* Copy offsets for distance codes 0..29 */
        1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
        257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
        8193, 12289, 16385, 24577};
static ush const cpdext[] = {    /* Extra bits for distance codes */
        0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,
        7, 7, 8, 8, 9, 9, 10, 10, 11, 11,
        12, 12, 13, 13};



/* Macros for inflate() bit peeking and grabbing.
   The usage is:
   
        NEEDBITS(j1)
        x = b & mask_bits[j1];
        DUMPBITS(j2)

   where NEEDBITS makes sure that b has at least j1 bits in it, and
   DUMPBITS removes j2 bits from b.  The macros use the variable k
   for the number of bits in b.  Normally, b and k are register
   variables for speed, and are initialized at the beginning of a
   routine that uses these macros from a global bit buffer and count.
   Frequently j2==j1; but j2 < j1 can occur when consecutive NEEDBITS
   are correlated, if the most significant bits of x agree with the
   least significant bits of the next x.  In such cases the bits of
   two x values overlap in the bit buffer, and the same bits in the
   compressed input contribute to two [or more] values of x.
   j2 > j1 is an error which is undetected.

   If we assume that EOB will be the longest code, then we will never
   ask for bits with NEEDBITS that are beyond the end of the stream.
   So, NEEDBITS should not read any more bytes than are needed to
   meet the request.  Then no bytes need to be "returned" to the buffer
   at the end of the last block.

   However, this assumption is not true for fixed blocks--the EOB code
   is 7 bits, but the other literal/length codes can be 8 or 9 bits.
   (The EOB code is shorter than other codes because fixed blocks are
   generally short.  So, while a block always has an EOB, many other
   literal/length codes have a significantly lower probability of
   showing up at all.)  However, by making the first table have a
   lookup of seven bits, the EOB code will be found in that first
   lookup, and so will not require that too many bits be pulled from
   the stream.
 */

#ifdef CRYPT
  uch cc;
#  define NEXTBYTE() \
     (decrypt ? (cc = get_byte(), zdecode(cc), cc) : get_byte())
#else
#  define NEXTBYTE()  (uch)get_byte()
#endif
#define NEEDBITS(n) {while(k<(n)){b|=((ulg)NEXTBYTE())<<k;k+=8;}}
#define DUMPBITS(n) {b>>=(n);k-=(n);}


/*
   Huffman code decoding is performed using a multi-level table lookup.
   The fastest way to decode is to simply build a lookup table whose
   size is determined by the longest code.  However, the time it takes
   to build this table can also be a factor if the data being decoded
   is not very long.  The most common codes are necessarily the
   shortest codes, so those codes dominate the decoding time, and hence
   the speed.  The idea is you can have a shorter table that decodes the
   shorter, more probable codes, and then point to subsidiary tables for
   the longer codes.  The time it costs to decode the longer codes is
   then traded against the time it takes to make longer tables.

   This results of this trade are in the variables lbits and dbits
   below.  lbits is the number of bits the first level table for literal/
   length codes can decode in one step, and dbits is the same thing for
   the distance codes.  Subsequent tables are also less than or equal to
   those sizes.  These values may be adjusted either when all of the
   codes are shorter than that, in which case the longest code length in
   bits is used, or when the shortest code is *longer* than the requested
   table size, in which case the length of the shortest code in bits is
   used.

   There are two different values for the two tables, since they code a
   different number of possibilities each.  The literal/length table
   codes 286 possible values, or in a flat code, a little over eight
   bits.  The distance table codes 30 possible values, or a little less
   than five bits, flat.  The optimum values for speed end up being
   about one bit more than those, so lbits is 8+1 and dbits is 5+1.
   The optimum values may differ though from machine to machine, and
   possibly even between compilers.  Your mileage may vary.
 */


int const lbits = 9;          /* bits in base literal/length lookup table */
int const dbits = 6;          /* bits in base distance lookup table */


/* If BMAX needs to be larger than 16, then h and x[] should be ulg. */
#define BMAX 16         /* maximum bit length of any code (16 for explode) */
#define N_MAX 288       /* maximum number of codes in any set */


struct huft *ha;               /* base of stack of struct huft */
static unsigned huft_next;     /* top-of-stack index */
static unsigned hufts;         /* track memory usage */

static
int
huft_free(hx)
huft_ndx hx;
{
  unsigned const hk = hx<<1;
  if (huft_next<=hk) {
    fprintf(stderr, "gzip: huft_free stack violation  %d %d\n", huft_next, hk);
    return 1;
  }
  ha = realloc(ha, hk * sizeof(struct huft));
  if (0==(huft_next = hk)) { /* stack became empty */
    ha = 0;  /* allow maximal freedom for next huft_alloc() */
  }
  return 0;
}

/* realloc() can move the block, so all existing pointers to previous
   allocations become invalid.  Remember only indices. */
static
huft_ndx
huft_alloc(n)
unsigned n;  /* number of struct huft to allocate */
{
  huft_ndx const rv = huft_next>>1;
  if (huft_next!=(rv<<1)) { /* huft_ndx overflow */
    return (huft_ndx)~0u;
  }
  n = (1+ n) & ~1;  /* round up to even */
  hufts += n;  /* track memory usage */
  huft_next += n;
  ha = realloc(ha, huft_next * sizeof(struct huft));
  if (0==ha) { /* realloc failed; out of memory */
    return (huft_ndx)~0u;
  }
  return rv;
}

int huft_build(b, n, s, d, e, t, m)
unsigned const *const b;/* code lengths in bits (all assumed <= BMAX) */
unsigned n;             /* number of codes (assumed <= N_MAX) */
unsigned const s;       /* number of simple-valued codes (0..s-1) */
ush const *const d;     /* list of base values for non-simple codes */
ush const *const e;     /* list of extra bits for non-simple codes */
huft_ndx *const t;      /* result: starting table */
int *const m;           /* maximum lookup bits, returns actual */
/* Given a list of code lengths and a maximum table size, make a set of
   tables to decode that set of codes.  Return zero on success, one if
   the given code set is incomplete (the tables are still built in this
   case), two if the input is invalid (all zero length codes or an
   oversubscribed set of lengths), and three if not enough memory. */
/* The tables of the set are linked through huft[-1].v.t, in order
   of allocation during this call. */
{
  unsigned a;                   /* counter for codes of length k */
  unsigned c[BMAX+1];           /* bit length count table */
  unsigned f;                   /* i repeats in table every f entries */
  int g;                        /* maximum code length */
  int h;                        /* table level */
  register unsigned i;          /* counter, current code */
  register unsigned j;          /* counter */
  register int k;               /* number of bits in current code */
  int l;                        /* bits per table (returned in m) */
  register unsigned const *p;   /* pointer into c[], b[], or v[] */
  register struct huft *q;      /* points to current table */
  struct huft r;                /* table entry for structure assignment */
  huft_ndx u[BMAX];             /* table stack */
  unsigned v[N_MAX];            /* values in order of bit length */
  register int w;               /* bits before this table == (l * h) */
  unsigned x[BMAX+1];           /* bit offsets, then code stack */
  unsigned *xp;                 /* pointer into x */
  int y;                        /* number of dummy codes added */
  unsigned z;                   /* number of entries in current table */


  /* Generate counts for each bit length */
  memzero(c, sizeof(c));
  p = b;  i = n;
  do {
    Tracecv(*p, (stderr, (n-i >= ' ' && n-i <= '~' ? "%c %d\n" : "0x%x %d\n"), 
	    n-i, *p));
    c[*p]++;                    /* assume all entries <= BMAX */
    p++;                      /* Can't combine with above line (Solaris bug) */
  } while (--i);
  if (c[0] == n)                /* null input--all zero length codes */
  {
    *t = (huft_ndx)~0u;
    *m = 0;
    return 0;
  }
  *t = huft_next>>1;


  /* Find minimum and maximum length, bound *m by those */
  l = *m;
  for (j = 1; j <= BMAX; j++)
    if (c[j])
      break;
  k = j;                        /* minimum code length */
  if ((unsigned)l < j)
    l = j;
  for (i = BMAX; i; i--)
    if (c[i])
      break;
  g = i;                        /* maximum code length */
  if ((unsigned)l > i)
    l = i;
  *m = l;


  /* Adjust last length count to fill out codes, if needed */
  for (y = 1 << j; j < i; j++, y <<= 1)
    if ((y -= c[j]) < 0)
      return 2;                 /* bad input: more codes than bits */
  if ((y -= c[i]) < 0)
    return 2;
  c[i] += y;


  /* Generate starting offsets into the value table for each length */
  x[1] = j = 0;
  p = c + 1;  xp = x + 2;
  while (--i) {                 /* note that i == g from above */
    *xp++ = (j += *p++);
  }


  /* Make a table of values in order of bit lengths */
  p = b;  i = 0;
  do {
    if ((j = *p++) != 0)
      v[x[j]++] = i;
  } while (++i < n);
  n = x[g];                   /* set n to length of v */


  /* Generate the Huffman codes and for each, make the table entries */
  x[0] = i = 0;                 /* first Huffman code is zero */
  p = v;                        /* grab values in bit order */
  h = -1;                       /* no tables yet--level -1 */
  w = -l;                       /* bits decoded == (l * h) */
  u[0] = 0;                     /* just to keep compilers happy */
  q = (struct huft *)NULL;      /* ditto */
  z = 0;                        /* ditto */

  /* go through the bit lengths (k already is bits in shortest code) */
  for (; k <= g; k++)
  {
    a = c[k];
    while (a--)
    {
      /* here i is the Huffman code of length k bits for value *p */
      /* make tables up to required level */
      while (k > w + l)
      {
        h++;
        w += l;                 /* previous table always l bits */

        /* compute minimum size table less than or equal to l bits */
        z = (z = g - w) > (unsigned)l ? l : z;  /* upper limit on table size */
        if ((f = 1 << (j = k - w)) > a + 1)     /* try a k-w bit table */
        {                       /* too few codes for k-w bit table */
          f -= a + 1;           /* deduct codes from patterns left */
          xp = c + k;
	  if (j < z)
	    while (++j < z)       /* try smaller tables up to z bits */
	    {
	      if ((f <<= 1) <= *++xp)
		break;            /* enough codes to use up j bits */
	      f -= *xp;           /* else deduct codes from patterns */
	    }
        }
        z = 1 << j;             /* table entries for j-bit table */

        /* allocate and remember new table */
        if ((u[h] = huft_alloc(z)) == (huft_ndx)~0u)
        {
          if (h)
            huft_free(u[0]);
          return 3;             /* not enough memory */
        }
	q = &ha[u[h]<<1];

        /* connect to last table, if there is one */
        if (h)
        {
          x[h] = i;             /* save pattern for backing up */
          r.b = (uch)l;         /* bits to dump before this table */
          r.e = (uch)(16 + j);  /* bits in this table */
          r.v.t = u[h];         /* huft_ndx of this table */
          j = i >> (w - l);     /* (get around Turbo C bug) */
          (&ha[u[h-1]<<1])[j] = r;        /* connect to last table */
        }
      }

      /* set up table entry in r */
      r.b = (uch)(k - w);
      if (p >= v + n)
        r.e = 99;               /* out of values--invalid code */
      else if (*p < s)
      {
        r.e = (uch)(*p < 256 ? 16 : 15);    /* 256 is end-of-block code */
        r.v.n = (ush)(*p);             /* simple code is just the value */
	p++;                           /* one compiler does not like *p++ */
      }
      else
      {
        r.e = (uch)e[*p - s];   /* non-simple--look up in lists */
        r.v.n = d[*p++ - s];
      }

      /* fill code-like entries with r */
      f = 1 << (k - w);
      for (j = i >> w; j < z; j += f)
        q[j] = r;

      /* backwards increment the k-bit code i */
      for (j = 1 << (k - 1); i & j; j >>= 1)
        i ^= j;
      i ^= j;

      /* backup over finished tables */
      while ((i & ((1 << w) - 1)) != x[h])
      {
        h--;                    /* don't need to update q */
        w -= l;
      }
    }
  }


  /* Return true (1) if we were given an incomplete table */
  return y != 0 && g != 1;
}


#if 0  /*{  asm_inflate.S equivalent */
int inflate_codes(
	struct huft const *const tl,  /* literal/length decoder table */
	struct huft const *const td,  /* distance decoder table */
	int const bl,  /* number of bits decoded by tl */
	int const bd   /* number of bits decoded by td[] */
)
/* inflate (decompress) the codes in a deflated (compressed) block.
   Return an error code or zero if it all goes ok. */
{
  unsigned r_crc = crc;
  unsigned w = wp;      /* current window position */

  /* inflate the coded data */
  for (;;)                      /* do until end of block */
  {
    struct huft const *t = tl;
    register unsigned e = bl;
    goto enter1;
    do {
      if (e == 99)
        return 1;
      e -= 16;
      t = &ha[t->v.t<<1];
enter1:
      e = (t += peekbits(e))->e;
      k += t->b;
    } while (e > 16);
    if (e == 16)                /* then it's a literal */
    {
      uch x;
      *(char volatile *)&slide[w];  /* tend PentiumPlain cache */
      x = (uch)t->v.n;
      slide[w++] = x;
      r_crc = crc_32_tab[((int)r_crc ^ x) & 0xff] ^ (r_crc >> 8);
      Tracevv((stderr, "%c", slide[w-1]));
      if (w == WSIZE)
      {
        flush_output(w);
        w = 0;
      }
    }
    else if (e != 15)    /* then it's a block to copy */
    {
      /* get length of block to copy */
      unsigned n = t->v.n + peekbits(e);
      k += e;

      /* decode distance of block to copy */
      t = td;
      e = bd;
      goto enter2;
      do {
        if (e == 99)
          return 1;
        e -= 16;
        t = &ha[t->v.t<<1];
enter2:
        e = (t += peekbits(e))->e;
        k += t->b;
      } while (e > 16);
     {
      /*  t  is dead after initializing  d  */
      unsigned d = w - t->v.n - peekbits(e);  /* index for copy */
      k += e;
      Tracevv((stderr,"\\[%d,%d]", w-d, n));

      /* do the copy */
      do {
        n -= (e = (e = WSIZE - ((d &= WSIZE-1) > w ? d : w)) > n ? n : e);
/* Heuristically:  e <= 7  half of the time.
   Inlining updcrc() gives more benefit than optimized memcpy.
*/
#if 0 && !defined(NOMEMCPY) && !defined(DEBUG)
        if (8 <= e && w - d >= e) /* (this test assumes unsigned comparison) */
        {
          updcrc(slide + d, e);
          memcpy(slide + w, slide + d, e);
          w += e;
          d += e;
        }
        else                      /* do it slow to avoid memcpy() overlap */
#endif /* !NOMEMCPY */
          do {
            uch const x = slide[d++];
            slide[w++] = x;
            r_crc = crc_32_tab[((int)r_crc ^ x) & 0xff] ^ (r_crc >> 8);
	    Tracevv((stderr, "%c", x));
            *(char volatile *)&slide[w];  /* tend PentiumPlain cache */
          } while (--e);
        if (w == WSIZE)
        {
          flush_output(w);
          w = 0;
        }
      } while (n);
     }
    }
    else  /* e == 15 */
      break;  /* EOB (end of block) */
  }


  /* restore the globals from the locals */
  crc = r_crc;
  wp = w;                       /* restore global window pointer */

  /* done */
  return 0;
}
#endif  /*}*/


int inflate_stored()
/* "decompress" an inflated type 0 (stored) block. */
{
  unsigned n;           /* number of bytes in block */
  unsigned w;           /* current window position */
  register ulg c = crc; /* temporary variable */

  /* make local copies of globals */
  w = wp;                       /* initialize window position */

  /* go to byte boundary */
  k += 7& -k;

  /* get the length and its complement */
  n = getbits(16);
  if (n != (unsigned)((~getbits(16)) & 0xffff))
    return 1;                   /* error in compressed data */


  /* read and output the compressed data */
  while (n--)
  {
    register uch const x = (uch)getbits(8);
    slide[w++] = x;
    c = crc_32_tab[((int)c ^ x) & 0xff] ^ (c >> 8);
    if (w == WSIZE)
    {
      flush_output(w);
      w = 0;
    }
  }
  crc = c;


  /* restore the globals from the locals */
  wp = w;                       /* restore global window pointer */
  return 0;
}



int inflate_fixed()
/* decompress an inflated type 1 (fixed Huffman codes) block.  We should
   either replace this with a custom decoder, or at least precompute the
   Huffman tables. */
{
  int i;                /* temporary variable */
  huft_ndx tlx;         /* literal/length code table */
  huft_ndx tdx;         /* distance code table */
  int bl;               /* lookup bits for tl */
  int bd;               /* lookup bits for td */
  unsigned l[288];      /* length list for huft_build */


  /* set up literal table */
  for (i = 0; i < 144; i++)
    l[i] = 8;
  for (; i < 256; i++)
    l[i] = 9;
  for (; i < 280; i++)
    l[i] = 7;
  for (; i < 288; i++)          /* make a complete, but wrong code set */
    l[i] = 8;
  bl = 7;
  if ((i = huft_build(l, 288, 257, cplens, cplext, &tlx, &bl)) != 0)
    return i;


  /* set up distance table */
  for (i = 0; i < 30; i++)      /* make an incomplete code set */
    l[i] = 5;
  bd = 5;
  if ((i = huft_build(l, 30, 0, cpdist, cpdext, &tdx, &bd)) > 1)
  {
    huft_free(tlx);
    return i;
  }


  /* decompress until an end-of-block code */
  if (inflate_codes(&ha[tlx<<1], &ha[tdx<<1], bl, bd))
    return 1;


  /* free the decoding tables in stack-like order, return */
  huft_free(tdx);
  huft_free(tlx);
  return 0;
}



int inflate_dynamic()
/* decompress an inflated type 2 (dynamic Huffman codes) block. */
{
  int i;                /* temporary variables */
  unsigned j;
  unsigned l;           /* last length */
  unsigned n;           /* number of lengths to get */
  huft_ndx tlx;
  struct huft *tl;      /* literal/length code table */
  huft_ndx tdx;
  struct huft *td;      /* distance code table */
  int bl;               /* lookup bits for tl */
  int bd;               /* lookup bits for td */
  unsigned nb;          /* number of bit length codes */
  unsigned nl;          /* number of literal/length codes */
  unsigned nd;          /* number of distance codes */
#ifdef PKZIP_BUG_WORKAROUND
  unsigned ll[288+32];  /* literal/length and distance code lengths */
#else
  unsigned ll[286+30];  /* literal/length and distance code lengths */
#endif

  /* read in table lengths */
  nl = 257 + getbits(5);      /* number of literal/length codes */
  nd = 1 + getbits(5);        /* number of distance codes */
  nb = 4 + getbits(4);         /* number of bit length codes */
#ifdef PKZIP_BUG_WORKAROUND
  if (nl > 288 || nd > 32)
#else
  if (nl > 286 || nd > 30)
#endif
    return 1;                   /* bad lengths */


  /* read in bit-length-code lengths */
  for (j = 0; j < nb; j++)
  {
    ll[border[j]] = getbits(3);
  }
  for (; j < 19; j++)
    ll[border[j]] = 0;


  /* build decoding table for trees--single level, 7 bit lookup */
  bl = 7;
  if ((i = huft_build(ll, 19, 19, NULL, NULL, &tlx, &bl)) != 0)
  {
    if (i == 1)
      huft_free(tlx);
    return i;                   /* incomplete code set */
  }

  if (tlx == (huft_ndx)~0u)	/* Grrrhhh */
	return 2;
  tl = &ha[tlx<<1];

  /* read in literal and distance code lengths */
  n = nl + nd;
  i = l = 0;
  while ((unsigned)i < n)
  {
    td = tl + peekbits(bl);
    k += td->b;
    j = td->v.n;
    if (j < 16)                 /* length of code in bits (0..15) */
      ll[i++] = l = j;          /* save last length in l */
    else if (j == 16)           /* repeat last length 3 to 6 times */
    {
      j = 3 + getbits(2);
      if ((unsigned)i + j > n)
        return 1;
      while (j--)
        ll[i++] = l;
    }
    else if (j == 17)           /* 3 to 10 zero length codes */
    {
      j = 3 + getbits(3);
      if ((unsigned)i + j > n)
        return 1;
      while (j--)
        ll[i++] = 0;
      l = 0;
    }
    else                        /* j == 18: 11 to 138 zero length codes */
    {
      j = 11 + getbits(7);
      if ((unsigned)i + j > n)
        return 1;
      while (j--)
        ll[i++] = 0;
      l = 0;
    }
  }


  /* free decoding table for trees */
  huft_free(tlx);


  /* build the decoding tables for literal/length and distance codes */
  bl = lbits;
  if ((i = huft_build(ll, nl, 257, cplens, cplext, &tlx, &bl)) != 0)
  {
    if (i == 1) {
      fprintf(stderr, " incomplete literal tree\n");
      huft_free(tlx);
    }
    return i;                   /* incomplete code set */
  }
  bd = dbits;
  if ((i = huft_build(ll + nl, nd, 0, cpdist, cpdext, &tdx, &bd)) != 0)
  {
    if (i == 1) {
      fprintf(stderr, " incomplete distance tree\n");
#ifdef PKZIP_BUG_WORKAROUND
      i = 0;
    }
#else
      huft_free(tdx);
    }
    huft_free(tlx);
    return i;                   /* incomplete code set */
#endif
  }


  /* decompress until an end-of-block code */
  if (inflate_codes(&ha[tlx<<1], &ha[tdx<<1], bl, bd))
    return 1;


  /* free the decoding tables in stack-like order, return */
  huft_free(tdx);
  huft_free(tlx);
  return 0;
}



int inflate_block(e)
int *e;                 /* last block flag */
/* decompress an inflated block */
{
  unsigned t;           /* block type */

  /* read in last block bit */
  *e = getbits(1);

  /* read in block type */
  t = getbits(2);

  /* inflate that block type */
  if (t == 2)
    return inflate_dynamic();
  if (t == 0)
    return inflate_stored();
  if (t == 1)
    return inflate_fixed();


  /* bad block type */
  return 2;
}



int inflate()
/* decompress an inflated entry */
{
  int e;                /* last block flag */
  int r;                /* result code */
  unsigned h;           /* maximum struct huft's malloc'ed */


  /* initialize window, bit buffer */
  wp = 0;
  in_end = &inbuf[insize];
  k = -(insize - inptr)<<3;

  /* decompress until the last block */
  h = 0;
  do {
    hufts = 0;
    if ((r = inflate_block(&e)) != 0)
      return r;
    if (hufts > h)
      h = hufts;
  } while (!e);

  k += 7& -k;  /* advance to byte boundary */
  inptr = &in_end[k>>3] - &inbuf[0];

  /* flush out slide */
  flush_output(wp);


  /* return success */
#ifdef DEBUG
  fprintf(stderr, "<%u> ", h);
#endif /* DEBUG */
  return 0;
}
